#ifndef __BrashCommands_h_included_p
#define __BrashCommands_h_included_p
//
// Copyright 2002-2016, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

#define BRASH_VERSION_STRING "01.01.40"

// version log:  
//
// 1.1.40 - Bug fix:      Bug in .linreg when last line of its input file did not end in \n
//                        corrected.
//
// 1.1.39 - Bug Fix:      segment violation crash when executing "cmd" /c <nothing>
//
// 1.1.38 - Enhancement:  adding -n and -z as test operators.  clarifying builtins logging.
//
// 1.1.37 - Enhancement:  Added additional test operators and modified builtin logging to 
//                        explain same.
//
// 1.1.36 - bug fixed:    Calling a function was superseeding the $@ variable in the current scope.
//
// 1.1.35 - bug fixed:    getopts bug if user specified multiple options in the same -group
//
// 1.1.34a - Port:        Port to Visual Studio 2015 (compiler rev 1900)
//
// 1.1.33 - Bug fixed:    $_ was not working when the previous command was a brash function. 
//          Bug fixed:    function definitions would not accept ; as a separator before the {
// 
// 1.1.32 - Bug fixed:    yet another cursor position bug when you scroll off the bottom
//                        of the windows console "screen buffer" while typing a command.
// 1.1.31 - Bug fixed:    Cursor position bug when the cursor is at the bottom of the
//                        scroll window.
//
// 1.1.30 - maintenance:  Code restructuring for pipe problems.
//
// 1.1.29 - Bug fixed:    Major bug fix -- Early while loop termination bug:
//                        When the $() or `` operators were used
//                        to invoke child processes in while-loop bodies when data was being
//                        piped into the loop.
//
// 1.1.28 - Bug fixed:    support for '\' in PS1 values.
//
// 1.1.27 - Bug fixed:    kind of fixed.  In pipe statements, if a downstream
//                        partner died quickly, then hangs would occur because
//                        upstream partners were unable to latch on the the
//                        input handle of the child before it completely vanished
//                        from running data structures.  This intermediate state
//                        made it impossible to pass the cancelWrite() operation
//                        to the upstream pipe when the child dies -- so the pipe
//                        would hang.  Increasing the time out before the upstream
//                        pipe attempts to connect gives the child a chance to die
//                        gracefully and disappear.  But it slows down all pipe 
//                        operations.  There seems to be no way for windows to
//                        safely synchronize in this case -- so a timeout is all
//                        that can be found to do.
// 1.1.26 - Improvement:  error messages for break and continue used out of place
// 1.1.25 - Bug Fixed:    the following commands no longer crash: 
//                           -  echo yes | exit 1
//                           -  eval exit
//                        However, eval exit does nothing now.  fix this later.
//          Bug Fixed:    child shell script error code is no longer lost.
// 1.1.24 - Improvement:  .wc now supports reading from stdin
// 1.1.23 - Improvement:  improved help listing from the set command
// 1.1.22 - Improvement:  added .stopwatch
// 1.1.21 - Improvement:  added -f and -h options to the set command
//
// 1.1.20 - Improvement:  added .linreg -- linear regression calculator.
//
// 1.1.19 - Improvement:  adding .solve command.
//          Bug fixed:  some error codes were being lost when you used command line
//          redirection.
//
// 1.1.18 - Bug fix:  When executing strings as commands, as when you do `cmd`
//          there were occaisional exceptions thrown due to closed stdin files.
//          Have removed same. 
//
// 1.1.17 - Fixed a major bug:  Redirecting stdout to a file inside a while loop
//          body was causing the close of stdin to the while loop as a whole.
//          this is a big no no because it was causing the while loop to terminate
//          early.  A pipe writing to the loop would get a pipe-closed error and 
//          would naturally terminate early.
//
// 1.1.16 - Improved the build.bat script so you don't need to set your path
//          to include the compiler.
//
//        - Added -i and -e options .regex
//
//        - Fixed a bug in the command line redirection of 1>&3 so that it works now.
//          Although passing standard handles 3 and higher to child processes does not
//          work.
//
//        - Reworked the IFS interpretation to fix a problem where:  for f in `cat file` did not
//          separate based on all words in the file, but used \n as the separator and produced
//          empty strings for \r and \t.  
//
// 1.1.15 - Fixed a bug in the way that the -script option works:  command line parameters after the script
//          name were not being passed into the top level environment
//
// 1.1.14 - Fixed a major bug in while statement handling:  the continue command was being treated like the
//          break command.  Added .rename
//
// 1.1.13 - Fixed a minor bug in the handling of the "--" option in the "echo" command.  The symptom was that
//          options that began with -- but had other stuff in them were being treated as if they were just "--"
//
//          Added .wc command
//
// 1.1.12 - Minor tweaks to enable detection of ^C in .ls .  Found a bug in the $( ... ) syntax.  Pipes blow its mind
//
// 1.1.11 - Added copyright information to the .ver builtin command
//
// 1.1.10 - Bug fix to .diff output and various builtins command message improvements.
//
// 1.1.9 -- Minor spelling corrections and improvements to .diff
//
// 1.1.8 -- Removed extraneous quotes around commands sent to external processes.
//
// 1.1.7 -- Added .expand.  Fixed a bug in the handling of errors during the
//          file redirection process when used in pipes.  The bug caused the
//          interpreter to hang rather than print the error and quit.
//          Note -- use -u to un-expand.
//
// 1.1.6 -- Added .head and .cut command
//
// 1.1.5 -- fixed a bug in piping from one external program to another and added
//          the "-rc file" option.
//
// 1.1.4 -- fixing writeBytes() end of line sequences so that map properly for 
//          notepad.
//
// 1.1.3 -- fixed bug in the shift command.  it now accepts expressions, not jsut
//          string constants.
//
// 1.1.2 -- fixed bug in the tab-completion logic so that
//          it understands ~ at the begining of the file name to mean %USERPROFILE%.
//          Also changed the opening banner to emphasize that this is a windows
//          specific clone.
// 

#include <cxxtls/scriptCommand.h>
#include "BrashCommandLine.h"
#include "BrashException.h"
#include  "BrashRedirection.h"
#include <cxxtls/AutoDestroyingPointer.h>
#include <cxxtls/threadMonitoredSet.h>
#include <cxxtls/ScriptLineStream.h>
#include <cxxtls/threadGate.h>
#include <cxxtls/threadSafeHandle.h>
#include <cxxtls/stringexpressionparser.h>
#include <sstream>

///@file  
///   Define commands executed for the Brash shell.  Note that "commands" are not
///   Brash statements.  
///
///   "Commands" are activities that the multithreaded execution
///   sysem executes.  Brash language statements occur inside there commands but they
///   are far from idential.

using namespace std;
using namespace cxxtls::ScriptFunctions;
using namespace cxxtls;



//==========================================================================================================
//==========================================================================================================
//==========================================================================================================

// typedef cxxtls::ScriptFunctions::StatementReturnCode StatementReturnCode;

//==========================================================================================================
//==========================================================================================================
//==========================================================================================================

bool consoleStillRunning();                 ///< return true if the user has not ^C the app while the console is up
                                            ///  @note  thread safe
void setConsoleRunFlag(bool keepRunning);   ///< Set the console running flag.  Used only by console ^c handler
                                            ///  @note thread safe.

//==========================================================================================================
//==========================================================================================================
//==========================================================================================================

// note:  There is no BourneLetStatement, "let" is handled as a "builtin" command and it evaluates its
//        parameters as arithmetic expressions.  

//==========================================================================================================
//==========================================================================================================
//==========================================================================================================

struct ProgramStructureException
    /// an exception that just means that we should shut down the current
    /// command or statementlist because something bad happened at runtime.
    ///
    /// This might be as simple as cd failing, or the use of an invalid operator in a test
    /// expression:   if [ -glarf ]
{
};

//==========================================================================================================
//==========================================================================================================
//==========================================================================================================


struct BrashThreadExitType
   // the exit status of a running thread.  This is typically only used by the console reader
{
    int exitCode_;

    std::string                currentDirectory_;   ///<  empty unless the "console" is involved
    std::vector<std::string>   arguments_;          ///<  empty unless the console is involved
    std::string                lastParm_;           ///<  last parm of last command 
    std::string                dirStack_;           ///<  dirstack from child
    std::string                IFS_;                ///<  IFS variable from child
    bool                       expandingNames_;     ///<  globbing() flag from ExecutionContext

    BrashThreadExitType(int exitCode, 
                        string const &currentDirectory, 
                        std::vector<std::string> const &args, 
                        std::string const &lastParm,
                        std::string const &dirStack,
                        std::string const &IFS,
                        bool               expandingNames
                       )
    :  exitCode_(exitCode)
    {
        // this code ensures that no two string involved point to the same reference counted
        // string body.

        currentDirectory_.assign(currentDirectory.begin(), currentDirectory.end());

        CXXTLS_FOREACH(std::string const &cur, args)
        {
           arguments_.push_back(std::string());
           arguments_.back().assign(cur.begin(), cur.end());
           
           // at this point, the arguments_ member strings have been copy constructed from an empty string into the 
           // copy stored in the vector.  However, before this loop completes, the temporary copy gets destructed.
        }

        lastParm_.resize(0);
        lastParm_.assign(lastParm.begin(), lastParm.end());

        dirStack_.resize(0);
        dirStack_.assign(dirStack.begin(), dirStack.end());

        IFS_.resize(0);
        IFS_.assign(IFS.begin(), IFS.end());

        expandingNames_ = expandingNames; 

    }

    BrashThreadExitType(int exitCode)
    :  exitCode_(exitCode)
    ,  expandingNames_(true)
    {
    }

    BrashThreadExitType()
    :  exitCode_(-1)
    ,  expandingNames_(true)
    {
    }


    static BrashThreadExitType unset;
    static BrashThreadExitType shuttingDown;


};

//==========================================================================================================
//==========================================================================================================
//==========================================================================================================

struct JobDefinition
{
    int jobId_;  // the only thing that matters for comparisons

    string args_;   

    ~JobDefinition()
    {
        jobId_ = 0;  // just a place for the debugger to stop
    }


    JobDefinition(int jobId)
    :  jobId_(jobId)
    {
    }

    JobDefinition(int jobid, string const &args)
    :  jobId_(jobid)
    ,  args_()
    {
        args_.assign(args.begin(), args.end());  // make args_ unique across all strings in all threads
                                                 // to prevent mult-threading issues
    }

    bool operator < (JobDefinition const &rhs) const
    {
       return jobId_ < rhs.jobId_;   // args are not considered.
    }


    JobDefinition(JobDefinition const &rhs)
    {
        jobId_ = rhs.jobId_;
        args_.assign(rhs.args_.begin(), rhs.args_.end());  // make args_ a new, unique across all strings text body
    }

    template<class Stream>
    friend Stream &operator<< (Stream &s, JobDefinition const &j)
    {
       return s << "BrashJob(" << j.jobId_ << ", " << j.args_ << ")";
    }


};

//==========================================================================================================
//==========================================================================================================
//==========================================================================================================


struct BrashScriptCommandHandler
:  public  cxxtls::ScriptCommandHandler
,  public  cxxtls::ScriptFunctions::JobController
   ///  This is the command handler for all commands executed by Brash.
   ///  Note that the all have an exit code.
   ///
   ///   "Commands" are activities that the multithreaded execution
   ///   sysem executes.  Brash language statements occur inside there commands but they
   ///   are far from idential.
   ///
   ///  A command handler consists of two parts:
   ///  @li  a thread safe queue into which commands can be pushed
   ///  @li  a pool of threads all of which are competing to get to execute commands as they
   ///       are pushed into the queue.
   ///  execute commands as they are
{
   BrashScriptCommandHandler(int threads)
   :  cxxtls::ScriptCommandHandler(threads)
   {
   }

   static void setExitCode(int value)
   {
      cxxtls::ScopedMutexLock lock(exitCodeGuard_);
      exitCode_ = value;
   }

   static int exitCode() 
   {
      cxxtls::ScopedMutexLock lock(exitCodeGuard_);
      return exitCode_;
   }

   static int newJobId()
   {
      cxxtls::ScopedMutexLock lock(exitCodeGuard_);
      ++jobId_;

      return jobId_;
   }

   static void registerJob(JobDefinition const &job)
   {
       CXXTLS_DEBUG("Adding job " << job);
       activeJobs_.add(job);
   }

   static BrashThreadExitType  waitForJob(JobDefinition const &jobId)
   {
      return activeJobs_.wait(jobId);
   }

   static void removeJob(JobDefinition const &jobId, BrashThreadExitType const &errorCode)
   {
       CXXTLS_DEBUG("Removing job " << jobId);
      activeJobs_.remove(jobId, errorCode);
   }

   static void shutdownJobs()
   {
       CXXTLS_DEBUG("Brash is shutting down");
      activeJobs_.shutdown();
   }

   static  void jobList(std::list<JobDefinition> &jobList)
   {
      activeJobs_.getList(jobList);
   }

   virtual StatementReturnCode executeDetached(ScriptThread::ThreadContext *executionContext,
                                               FunctionScope                *functionScope,
                                               JobController                *jobController,
                                               ThreadSafeHandle<Statement>  &statement
                                              );


private:
   static cxxtls::Mutex exitCodeGuard_;
   static int exitCode_;
   static int jobId_;
   static cxxtls::ThreadMonitoredSet<JobDefinition, BrashThreadExitType> activeJobs_;


};


//==========================================================================================================
//==========================================================================================================
//==========================================================================================================

struct exitCommand
:  public cxxtls::ScriptCommand
    ///  terminate brash's execution with a given exit code.  this class was used
    ///  in debug and does not reflect the brash "exit" statement.
{
   int exitCode_;

   exitCommand(BrashScriptCommandHandler* owner, int exitCode)
   :  cxxtls::ScriptCommand(owner)
   ,  exitCode_(exitCode)
   {
   }

   virtual void run()
   {
       BrashScriptCommandHandler::setExitCode(exitCode_);

       shutdown();
   }

}; 

//==========================================================================================================
//==========================================================================================================
//==========================================================================================================


struct ReadConsoleCommand
:  public cxxtls::ScriptCommand
    ///  Prompt the user for brash language input and execute it one line at a time.
    ///
    ///  this is a debug feature to be deleted
{
   
   cxxtls::StringExpr::ExecutionVariables env_;

   ReadConsoleCommand(BrashScriptCommandHandler* owner, cxxtls::StringExpr::ExecutionVariables const &env)
   :  cxxtls::ScriptCommand(owner)
   ,  env_(env)
   {
   }

   virtual void run();

}; 

//==========================================================================================================
//==========================================================================================================
//==========================================================================================================

struct ReadFileCommand
:  public cxxtls::ScriptCommand
    ///  Prompt the user for brash language input and execute it one line at a time.
{
   
   std::string                              filename_;
   cxxtls::StringExpr::ExecutionVariables   env_;
   std::stringstream                       *stringStream_;  // not owned by this class
   std::vector<std::string>                 args_;  // arg0 is filename or "text of command.  must have at least 1 arg (arg0) if we are reading a file not a string

   ReadFileCommand(std::string const &filename, BrashScriptCommandHandler* owner, cxxtls::StringExpr::ExecutionVariables const &env, std::vector<std::string> const &args)
   : cxxtls::ScriptCommand(owner)
   , filename_(filename)
   ,  env_(env)
   ,  stringStream_(0)
   ,  args_(args)
   {
   }

   ReadFileCommand(std::stringstream *stringStream, std::string const &command, BrashScriptCommandHandler* owner, cxxtls::StringExpr::ExecutionVariables const &env)
   : cxxtls::ScriptCommand(owner)
   , filename_(command)
   ,  env_(env)
   ,  stringStream_(stringStream)  // not owned by this code
   {
   }


   virtual void run();

}; 

//==========================================================================================================
//==========================================================================================================
//==========================================================================================================




struct BrashStatementList
:  public cxxtls::ScriptCommand
    /// A command that will execute a list of fully self contained BrashStatements.  
    /// That is:  This class encapsulates a list of statements
    /// which have their own thread safe copy of the environment, command line arguments,
    /// file handles, etc.
    ///
    /// Note:   see setSuppressExit(), below.
    ///
    /// Normally, if any command in the list is "exit", then brash executable will shutdown
{
   cxxtls::ScriptThread::ThreadContext*                   context_;     // can't be an auto-ptr, since it must be destroyed, not deleted
   std::auto_ptr<cxxtls::ScriptFunctions::StatementList>  statements_;
   int                                                    jobId_;
   bool                                                   startedByConsole_;
   bool                                                   suppressExit_;   // flag means:  don't shutdown on exit of this job

   BrashStatementList(BrashScriptCommandHandler *owner, 
                   cxxtls::ScriptThread::ThreadContext    *myContext,
                   cxxtls::ScriptFunctions::StatementList *statements,
                   int                                     jobId,
                   bool                                    startedByConsole=false
                  )
   :  cxxtls::ScriptCommand(owner)
   ,  context_(myContext)
   ,  statements_(statements)
   ,  jobId_(jobId)
   ,  startedByConsole_(startedByConsole)
   ,  suppressExit_(false)
   {
        CXXTLS_DEBUG("BrashStatementList(" << std::hex << (long )(this) << std::hex << ")");  
   }

   void setSuppressExit(bool v)
      ///<  Call this function in order to instruct the statement list to ignore exit calls -- treating them
      ///   more like function returns.  The only reason that you call this function is in the case where you are
      ///   executing an external command and you don't want its exit to terminate the rest of the jobs that running.
   {
      suppressExit_ = v;
   }

   virtual ~BrashStatementList()
   {
        CXXTLS_DEBUG("~BrashStatementList(" << std::hex << (long )(this) << std::hex << ")");  

        cxxtls::ScriptThread::ThreadContext::destroy(context_);  
   }



   virtual void run();
};

//==========================================================================================================
//==========================================================================================================
//==========================================================================================================

struct BrashPipeMember
:  public cxxtls::ScriptCommand
    /// A single statement (or statement list statement) with its environment and
    /// context set up for piping.
{
   cxxtls::ScriptThread::ThreadContext*                   context_;     // may or may not be owned, see deleteContext
   ThreadSafeHandle<Statement>                            statement_;  
   int                                                    jobId_;
   bool                                                   deleteContext_;  // flag indicating to delete the context_ or not.
   ThreadGate                                            &finishLine_;
   int                                                    reverseChainIndex_;  // 0 is the last item in the chain
   bool                                                   closeFilesOnExit_;

   int                                                    pipeDepth_;  // mainly a debugging feature, nested pipe statements
                                                                       // confuse the developers without some context

   bool                                                   terminateInputPipe_;

   BrashPipeMember(BrashScriptCommandHandler              *owner, 
                   cxxtls::ScriptThread::ThreadContext    *myContext,
                   ThreadSafeHandle<Statement>            &statement,
                   int                                     jobId,
                   bool                                    deleteContext,
                   ThreadGate                             &finishLine,
                   int                                     reverseChainIndex,
                   bool                                    closeFilesOnExit,
                   bool                                    terminateInputPipe
                  )
   :  cxxtls::ScriptCommand(owner)
   ,  context_(myContext)
   ,  statement_(statement)
   ,  jobId_(jobId)
   ,  deleteContext_(deleteContext)
   ,  finishLine_(finishLine)
   ,  reverseChainIndex_(reverseChainIndex)
   ,  closeFilesOnExit_(closeFilesOnExit)
   ,  terminateInputPipe_(terminateInputPipe)
   {
        try { pipeDepth_ = lexical_cast<int>( myContext->variables().find("BRASH_PIPE_DEPTH") ); }
        catch(...)
        {
           pipeDepth_ = 10;   
        }

        CXXTLS_DEBUG("BrashPipeMember( depth=" << pipeDepth_ << ", reverseIndex=" << reverseChainIndex_ << "," <<  std::hex << (long )(this) << std::hex << ") ===============================");  
   }

   virtual ~BrashPipeMember()
   {
        CXXTLS_DEBUG("~BrashPipeMember(" << reverseChainIndex_ << "," <<  std::hex << (long )(this) << std::hex << ") ===============================");  

        #ifdef DebugBuild
            CXXTLS_LOG("BrashPipeMember: " << (void*)(this) << ", ends" );
        #endif

        finishLine_.signal();   // tell my owner I am done

        if(deleteContext_)
            cxxtls::ScriptThread::ThreadContext::destroy(context_);  
   }


   virtual void run();
};


//==========================================================================================================
//==========================================================================================================
//==========================================================================================================

struct BrashStatementListAndEnvironment
:  public BrashStatementList
   // same as a brash statement list but has its own copy of the environment which gets
   // deleted on destruction
{
    std::auto_ptr<cxxtls::StringExpr::ExecutionVariables> env_;

   virtual ~BrashStatementListAndEnvironment()
   {
      // automatically calls parent class' destructor
   }


   BrashStatementListAndEnvironment
                  (BrashScriptCommandHandler *owner, 
                   cxxtls::ScriptThread::ThreadContext    *myContext,
                   cxxtls::ScriptFunctions::StatementList *statements,
                   int                                     jobId,
                   cxxtls::StringExpr::ExecutionVariables *env
                  )
   : BrashStatementList(owner, myContext, statements, jobId)
   , env_(env)
   {
   }


};

//==========================================================================================================
//==========================================================================================================
//==========================================================================================================


struct BrashNoOpStatement
:  public cxxtls::ScriptFunctions::Statement
{
        virtual ~BrashNoOpStatement() {}

        virtual StatementReturnCode
        execute(  cxxtls::ScriptThread::ThreadContext          *executionContext,
                  cxxtls::ScriptFunctions::FunctionScope       *functionScope,
                  cxxtls::ScriptFunctions::JobController       *jobController,
                  bool                                          skipArg0            // for use by the command "builtin"
               ) const;
};

//==========================================================================================================
//==========================================================================================================
//==========================================================================================================

struct BrashBuiltinCommandExecutor
    ///  a "base class" from which built in commands derive.
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const = 0;

    virtual string abstract() const = 0;

    //==========================================================================================================
    //==========================================================================================================
    //==========================================================================================================


    struct WorkingHandlesAndEnvironmentPointer
        /// Either a simple pass through to the passed in executionContext, or a modified copy
        /// thereof.  The modified copy only exists if the BrashCommandLine object has either
        /// variable superseeds or redirection operators on the command line.
        ///
        /// The idea is that a brash builtin executor will use this class to make the needed determinations
        /// and clones and will use this class instead of the executioncontext passed into its execute function
    {
    private:
         AutoDestroyingPointer<cxxtls::ScriptThread::ThreadContext>       clonedContext_;        // possibly null, if no command line redirections

         cxxtls::ScriptThread::ThreadContext*                             workingContext_;       // original or clone
         cxxtls::ScriptThread::ThreadContext*                             originalContext_;      // original 

    public:
         
        ~WorkingHandlesAndEnvironmentPointer()
        {
            cxxtls::ScriptThread::ThreadContext *tmp = workingContext_;  // set a breakpoint here if needed.
            workingContext_ = tmp; 

            if(workingContext_ != originalContext_)
            {
               // some sort of superseed occurred

               originalContext_->variables().setErrorCode(workingContext_->variables().errorCode());  // make sure while loops terminate even if redirection occurs
               originalContext_->variables().setLastParmOfLastCommand(workingContext_->variables().lastParmOfLastCommand());

            }

        }

         WorkingHandlesAndEnvironmentPointer( BrashCommandLine                     * commandLine,
                                              cxxtls::ScriptThread::ThreadContext  *executionContext
                                            );

         cxxtls::ScriptThread::ManagedFileHandles* fileHandles() { return workingContext_->fileHandles_; }  ///< lets you treat *this as a pointer
                                                                                                          ///  to the file handles

         cxxtls::StringExpr::ExecutionContext&  variables() { return workingContext_->variables(); }
             ///<  get access to the modified enviornment variable set.  Modified by superseeds 

         cxxtls::ScriptThread::ThreadContext* executionContext() { return workingContext_; } ///< return original or modified copy


    private:
        // not implemented....
        WorkingHandlesAndEnvironmentPointer();
        WorkingHandlesAndEnvironmentPointer(WorkingHandlesAndEnvironmentPointer const &);
        WorkingHandlesAndEnvironmentPointer &operator=(WorkingHandlesAndEnvironmentPointer const &);
    };

    static
    void evaluateCommandLine(BrashCommandLine                     * commandLine,
                             cxxtls::ScriptThread::ThreadContext  * executionContext,
                             std::vector<std::string>             & args,
                             bool                                   skipArg0, // for use by the command, "builtin"
                             bool                                   globbing=true,
                             bool                                   splitting=true
                            );
        ///<  Evaluate the string expressions that are the command line and store the results in args


    //==========================================================================================================
    //==========================================================================================================
    //==========================================================================================================

    class CommandLineArguments
       ///  This class exists to populate its members, argc, and argv, with values that are
       ///  understandable as main(argc,aragv).  This is mainly so that the existing
       ///  getops() function can be used because using this class creates a slightly
       ///  unnecessary overhead -- you could just call evaluateCommandLine() directly and
       ///  use the vector of strings directly, bit if you want to use getops(), you have to
       ///  have the standard C defintion of argc and argv.  That is what this class does.
    {
        std::vector<char*>       argPointers_;  // contains the trailing null pointer as expected
 
    public:

        std::vector<std::string> args_;   // populate by the constructor
        int                      argc;    // ditto
        char* const*             argv;    // ditto.  You cannot modify this variable, or through it.  Modify argPointers_ instead!

        CommandLineArguments(BrashCommandLine                     * commandLine,
                             cxxtls::ScriptThread::ThreadContext  * executionContext,
                             bool                                   skipArg0,              // for use by command, "builtin"
                             bool                                   globbing=true,
                             bool                                   splitting=true
                            );

        void shift()
        {
            if(argc > 1)
            {
                args_.erase(args_.begin() + 1);
                --argc;
                for(size_t i = 0; i < args_.size(); ++i)
                    argPointers_[i] = (char*)args_[i].c_str();

                argPointers_[args_.size()] = 0;
            }
        }

    };


};

//==========================================================================================================
//==========================================================================================================
//==========================================================================================================

struct BrashExternalCommandExecutor
:  public BrashBuiltinCommandExecutor
{
    ///
    ///   This class is responsible for executing commands that are not in some way built into the
    ///   program itself.  These fall into 3 basic categories:
    ///
    ///       @li  native binary executables
    ///       @li  scripts to be executed using CMD.EXE
    ///            <blockquote>
    ///                 @li BAT
    ///                 @li CMD
    ///            </blockquote>
    ///       @li  BRASH scripts (external scripts)
    ///
    ///   External brash scripts are executed internally -- that is, they are parsed and read and
    ///   executed by the current brash executables.
    ///
    ///   External commands can be specified by complete filename or by path lookup.  When path lookup is
    ///   used, there is either no directory component in the command name or if there is a directory
    ///   component, it is a relative component.   The PATH variable is searched in both of these cases.
    ///
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const;

    virtual std::string abstract() const;
};

//==========================================================================================================
//==========================================================================================================
//==========================================================================================================

extern BrashExternalCommandExecutor brashBuiltinExecuteExternalCommand;

extern BrashBuiltinCommandExecutor const *findBuiltin(std::string const &name);  // return a pointer to the executor object
                                                                                  // for a builtin command -- or return 0.
//==========================================================================================================
//==========================================================================================================
//==========================================================================================================


struct BrashCommandLineStatement
:  public cxxtls::ScriptFunctions::Statement
    ///  An object that represents a command-line-style statement in the Brash language.  This object,
    ///  when executed, executes a brash "command line".  Syntactically, command
    ///  line statements look like this:
    ///@begincode
    ///
    ///   [var1=value1 var2=value2 ...]  CommandNameOrPath   [arguments ...]   [FileRedirections]
    ///
    ///@endcode
    ///
    ///   Command line states can invoke
    ///   @li  built in commands
    ///   @li  shell functions
    ///   @li  other scripts
    ///   @li  program files
    ///
{
        std::auto_ptr<BrashCommandLine>      command_;

        DECLARE_CXXTLS_EXCEPTION(EmptyCommandException);

        BrashBuiltinCommandExecutor const *commandExecutor_;

        void detectExecutionClass();

        BrashCommandLineStatement(BrashCommandLine *cmd, bool maybeBuiltin)
        :  command_(cmd)
        ,  commandExecutor_(&brashBuiltinExecuteExternalCommand)
           ///  take ownership of and prepare to execute a given BrashCommandLine object.
        {
            if(maybeBuiltin)
                detectExecutionClass();   // populates commandExecutor_ with one of the builtin command type executors

            // note:  the execution of "external" commands is done via  an unnamed builtin command, 
            //        in case you are wondering.
        }

        string description() const;
           ///<  return the name of the command being executed

        virtual ~BrashCommandLineStatement() 
        {
            CXXTLS_HEAVYDEBUG("~BrashCommandLineStatement(" << std::hex << (long )(this) << ")");  
        }

        virtual StatementReturnCode 
        execute(  cxxtls::ScriptThread::ThreadContext          *executionContext,
                  cxxtls::ScriptFunctions::FunctionScope       *functionScope,
                  cxxtls::ScriptFunctions::JobController       *jobController,
                  bool                                          skipArg0            // for use by the command "builtin"
               ) const;
            // just invokes commandExecutor_

        virtual std::string getDebugDescription() const 
        {
            /// get a string roughly describing the statement

            string rv;

            if(command_.get())
            {
                if(command_->arguments_.get())
                {
                    int count = 0;

                    CXXTLS_FOREACH(BrashCommandLine::ARGUMENTS_LIST::value_type const &cur,
                                   *command_->arguments_
                                  )
                    {
                        ++count;

                        if(count == 1)
                        {
                            string value;

                            if(cur->isSimple(value))
                            {
                                rv += value;
                            }
                            else
                                rv += "command ?";
                        }
                        else
                           break;
                        
                    }
                }
                else
                   rv += "<bad cmd line format>";

            }
            else
               rv += "<bad cmd line format>";


            return rv;

        }

};

//==========================================================================================================
//==========================================================================================================
//==========================================================================================================

bool parseBrash(cxxtls::ScriptLineStream                 &stream,
                BrashStatementList                       &output,
                cxxtls::ScriptThread::ManagedFileHandles &messageFiles,
                cxxtls::ScriptThread::ThreadContext      *parsingContext,
                bool                                      wholeFileAsLine = false
               );
    ///<  Parse lines from the specified stream and insert them as commands
    ///   into the output statement list.  Write error messages to the
    ///   stdout stream and stderr stream specified by messageFiles.
    ///
    ///   @param wholeFileAsLine Causes the entire file to be read into a single line
    ///                          and parsed that way.  Newlines will be kept in the string.
    ///
    ///   @return true if the parsing was successful and the statement should be
    ///           executed.
    ///
    ///   @return false if the statements should just be deleted.


//==========================================================================================================
//==========================================================================================================
//==========================================================================================================

enum ParseStatementReturnCode
{
   PSRC_GOOD=1,  // parseStatement successfully parsed a statement
   PSRC_BAD =2,  // syntax error or other problem that did not cause an exception
   PSRC_EOL =3,  // end of list.  When you are processing {}, (), if, while, or for statements, this return code
                 // means "good" but we hit the end of the statement list:  }, ), done, or fi
};

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

struct EndOfStatementListDetector
    /// A configurable way to specify the end of a statement list.  Construct an object of this type and
    /// pass it to parseStatement.
{
   bool endsWithKeyword_;      // done, or fi
   bool endsWithCharacter_;    // ), }
   char endingCharacter_;      
   string endingKeyword_;
   string endingKeyword2_;

   bool isEndingKeyword(string const &word) const
   {
      return word == endingKeyword_ || word == endingKeyword2_;
   }

   EndOfStatementListDetector()
   :  endsWithKeyword_(false)
   ,  endsWithCharacter_(false)
       /// the default constructor produces a detector that ends with  end of file and nothing else.
   {
   }

   EndOfStatementListDetector(string endingKeyword)
   :  endsWithKeyword_(true)
   ,  endsWithCharacter_(false)
   ,  endingCharacter_(0)
   ,  endingKeyword_(endingKeyword)
   ,  endingKeyword2_(endingKeyword)
       /// when you construct with a string, you are saying that detector ends when the command line has a certain word on it.
   {
   }

   EndOfStatementListDetector(string endingKeyword, string endingKeyword2)
   :  endsWithKeyword_(true)
   ,  endsWithCharacter_(false)
   ,  endingCharacter_(0)
   ,  endingKeyword_(endingKeyword)
   ,  endingKeyword2_(endingKeyword2)
       /// when you construct with a string, you are saying that detector ends when the command line has a certain word on it.
   {
   }


   EndOfStatementListDetector(char endingCharacter)
   :  endsWithKeyword_(false)
   ,  endsWithCharacter_(true)
   ,  endingCharacter_(endingCharacter)
   ,  endingKeyword_()
       /// when you construct with a character, you are saying that the statement list ends when the first character of a
       /// command line begins with that character.
   {
   }

};

//==========================================================================================================
//==========================================================================================================
//==========================================================================================================

      ParseStatementReturnCode
                parseStatement(
                               string::const_iterator           &f,
                               string::const_iterator           &l,
                               cxxtls::ScriptLineStream         &stream,
                               cxxtls::ScriptFunctions::StatementListStatement  &output,
                               cxxtls::ScriptThread::ManagedFileHandles &messageFiles,
                               cxxtls::ScriptInputLine                  &command,
                               EndOfStatementListDetector const &endDetector,
                               cxxtls::ScriptThread::ThreadContext      *parsingContext
                              );
      ///<  Parse the contents of the specified command, using [f,l) as the boundaries of the 
      ///   "current input line line" to get a brash command.  If the command requires additional
      ///   input, use the specified stream to obtain it.  
      ///
      ///   The original command should be populated from the same stream using the handleLineContinuations
      ///   function, below.

//==========================================================================================================
//==========================================================================================================
//==========================================================================================================

bool handleLineContinuations(  cxxtls::ScriptLineStream  &stream,
                               cxxtls::ScriptInputLine   &command,
                               bool                       wholeFileAsLine = false
                              );
    /// Read text from the input stream and handle
    /// line continuations.  A line contination is a \ just before the end of 
    /// line.  This may or may not be easy for a given text editor to generate. 
    /// Lots of times, the user causes a syntax error because they don't put the
    /// line continuation in correctly, but the rules as what they are.  Return
    /// false on end of file or error.
    ///
    /// @param stream           The stream from which to read the data
    /// @param command          The location to store the read line
    /// @param wholeFileAsLine  Read the entire file as a single string (including embedded newlines)
    ///
    /// @note  To control the prompt that the user sees, the caller should use AutoPoppingPushPrompt,
    ///         if needed.


//==========================================================================================================
//==========================================================================================================
//==========================================================================================================



    struct BourneWhileStatement
    : public Statement
    {
        virtual ~BourneWhileStatement();

        ThreadSafeHandle<Statement>  condition_;
        ThreadSafeHandle<Statement>  loopBody_;

        std::auto_ptr<RedirType::REDIRECTION_LIST> redirections_;

        BourneWhileStatement(ThreadSafeHandle<Statement> &condition, ThreadSafeHandle<Statement> &loopBody)
        :  condition_(condition)
        ,  loopBody_(loopBody)
        {
        }

        virtual StatementReturnCode execute(ScriptThread::ThreadContext             *executionContext,
                                            FunctionScope                           *functionScope,
                                            cxxtls::ScriptFunctions::JobController  *jobController,
                                            bool                                          skipArg0            // for use by the command "builtin"
                                           ) const;

        void setRedirections(RedirType::REDIRECTION_LIST*p)
           ///  replace the existing redirections with a new pointer and take ownership of it.
        {
           redirections_.reset(p);
        }

        virtual std::string getDebugDescription() const 
        {
            return "while ... ";
        }


    };


//==========================================================================================================
//==========================================================================================================
//==========================================================================================================

    struct BourneBlockStatement
    : public Statement
    {
        virtual ~BourneBlockStatement();

        ThreadSafeHandle<Statement> block_;

        std::auto_ptr<RedirType::REDIRECTION_LIST> redirections_;

        BourneBlockStatement(ThreadSafeHandle<Statement> &block)
        :  block_(block)
        {
        }

        virtual StatementReturnCode execute(ScriptThread::ThreadContext             *executionContext,
                                            FunctionScope                           *functionScope,
                                            cxxtls::ScriptFunctions::JobController  *jobController,
                                            bool                                          skipArg0            // for use by the command "builtin"
                                           ) const;

        void setRedirections(RedirType::REDIRECTION_LIST*p)
           ///  replace the existing redirections with a new pointer and take ownership of it.
        {
           redirections_.reset(p);
        }

        virtual std::string getDebugDescription() const 
        {
            if(&*block_)
               return block_->getDebugDescription();
            else
               return " { <empty> } ";
        }


    };

//==========================================================================================================
//==========================================================================================================
//==========================================================================================================

    struct BourneIfStatement
    : public Statement
    {
        virtual ~BourneIfStatement();

        ThreadSafeHandle<Statement> condition_;
        ThreadSafeHandle<Statement> mainClause_;
        ThreadSafeHandle<Statement> elseClause_;

        std::auto_ptr<RedirType::REDIRECTION_LIST> redirections_;

        BourneIfStatement(ThreadSafeHandle<Statement> &condition, ThreadSafeHandle<Statement> &mainClause, ThreadSafeHandle<Statement> &elseClause)
        :  condition_(condition)
        ,  mainClause_(mainClause)
        ,  elseClause_(elseClause)
        {
        }

        virtual StatementReturnCode execute(ScriptThread::ThreadContext             *executionContext,
                                            FunctionScope                           *functionScope,
                                            cxxtls::ScriptFunctions::JobController  *jobController,
                                            bool                                          skipArg0            // for use by the command "builtin"
                                           ) const;

        void setRedirections(RedirType::REDIRECTION_LIST*p)
           ///  replace the existing redirections with a new pointer and take ownership of it.
        {
           redirections_.reset(p);
        }

        virtual std::string getDebugDescription() const 
        {
            return "if ... ";
        }


    };

//==========================================================================================================
//==========================================================================================================
//==========================================================================================================

    struct BourneForStatement
    :  public Statement
        // The execution time structure of either a "for" or a "select" statement
    {
        virtual ~BourneForStatement();

        std::auto_ptr<Statement>  loopStatement_;

        std::list<StringExpression*> valueList_;  // list of values over which to iterate

        std::auto_ptr<RedirType::REDIRECTION_LIST> redirections_;

        std::string varname_;

        bool isSelect_;

        bool isArithmetic_;  // if true, then the valueList_ contains arithmetic assignments or expressions
                             // that define the for loop values and the conditions for ending the loop, and
                             // incrementing the loop variable.
                             //
                             // This variable is only true when this syntax is used
                             //
                             //   for (( i=1;  i < 300; ++i ))
                             //   do
                             //     ...
                             //   done

        BourneForStatement(std::string const &varname, std::list<StringExpression*> &valueList, Statement *loopStatement, bool isSelect, bool isArithmetic)
        :  loopStatement_(loopStatement)
        ,  varname_(varname)
        ,  isSelect_(isSelect)
        ,  isArithmetic_(isArithmetic)
           /// construct a new Bourne style for statement representation.  Take ownership of the loopStatement
           /// passed in as well as the stringExpressions passed in as the valueList.
           ///
           /// @param[in]      varname           The name of the loop variable
           /// @param[in,out]  valueList         The list of string expressions that become the values iterated over
           /// @param[in]      loopStatement     The statement to execute on each selected string value
           /// @param[in]      isSelect          A flag that means this is a select statement, not a normal for statement
           ///
           /// @warning The valueList that you pass in will be emptied when you pass it to this constructor.
        {
           valueList_.swap(valueList);
        }

        virtual StatementReturnCode execute(ScriptThread::ThreadContext             *executionContext,
                                            FunctionScope                           *functionScope,
                                            cxxtls::ScriptFunctions::JobController  *jobController,
                                            bool                                          skipArg0            // for use by the command "builtin"
                                           ) const;

        void setRedirections(RedirType::REDIRECTION_LIST*p)
           ///  replace the existing redirections with a new pointer and take ownership of it.
        {
           redirections_.reset(p);
        }

        virtual std::string getDebugDescription() const 
        {
            return "for ... ";
        }

    };

//==========================================================================================================
//==========================================================================================================
//==========================================================================================================

    struct BournePipeStatement
    :  public Statement
    {
        virtual ~BournePipeStatement();

        mutable std::list< ThreadSafeHandle<Statement> > chain_;    // handles violate const so this is no worse than they are in general

        BournePipeStatement(std::list< ThreadSafeHandle<Statement> > &chain)
            /// construct a pipe statement from a list of statements that are piped together.
            /// The constructor empties the input list.
            ///
            /// @param[in,out] chain
        {
            chain_.swap(chain);
        }

        virtual StatementReturnCode execute(ScriptThread::ThreadContext             *executionContext,
                                            FunctionScope                           *functionScope,
                                            cxxtls::ScriptFunctions::JobController  *jobController,
                                            bool                                          skipArg0            // for use by the command "builtin"
                                           ) const;

        virtual std::string getDebugDescription() const 
        {
           string rv;

           if(chain_.empty())
              return "?|?";

           int count=0;

           CXXTLS_FOREACH(ThreadSafeHandle<Statement> &cur, chain_)
           {
              ++count;

              if(count > 1)
                  rv += " | ";

              rv += cur->getDebugDescription();
              
           }

           return rv;
        }

    };

//==========================================================================================================
//==========================================================================================================
//==========================================================================================================

extern map<string, BrashBuiltinCommandExecutor*> builtinCommands; // this is a constant after threading starts


extern
StatementReturnCode
brashReadAndExecuteInContext(std::string const &fileName,
                             cxxtls::ScriptThread::ThreadContext                 *executionContext_in,
                             cxxtls::StringExpr::ExecutionContext::ArgsVector    &args_in,
                             bool                                                 suppressExit  // treats exit statements like function returns
                            );
   // look for the source in brashBuiltins.cxx


struct Is_IFS_Char
  ///  If you want to use the parse_words method to split a string
  ///  into substrings, separated by some delimiter, use this functor.
{
    std::string list_;

    size_t size_;

    Is_IFS_Char(std::string const &s)
    :  list_(s)
    {
        size_ = list_.size();
    }

    bool operator() (char c) const 
    { 
        return list_.find_first_of(c) < size_;
    }
};

//==========================================================================================================
//==========================================================================================================
//==========================================================================================================

    struct InstallFunctionStatement
    :  public Statement
    {

        string name_;

        ThreadSafeHandle<Statement> body_;

        string file_;
        int line_;
        

        InstallFunctionStatement(string const &name, ThreadSafeHandle<Statement> &body, string const &file, int line)
        :  name_(name)
        ,  body_(body)
        ,  file_(file)
        ,  line_(line)
        {
        }

        virtual ~InstallFunctionStatement();

        virtual StatementReturnCode execute(ScriptThread::ThreadContext             *executionContext,
                                            FunctionScope                           *functionScope,
                                            cxxtls::ScriptFunctions::JobController  *jobController,
                                            bool                                          skipArg0            // for use by the command "builtin"
                                           ) const;

        std::string getDebugDescription() const
        {
           return string("install function") + name_;
        }
    };


//==========================================================================================================
//==========================================================================================================
//==========================================================================================================


struct CaseMember
{
    std::list<StringExpression*>*                 conditions_;
    std::auto_ptr<StatementListStatement>         statements_;

    CaseMember(std::list<StringExpression*>       *condition,
               StatementListStatement             *statements
              )
    :  conditions_(condition)
    ,  statements_(statements)
    {
    }

   ~CaseMember()
   {
      if (conditions_)
      CXXTLS_FOREACH(StringExpression *cur, *conditions_)
          delete cur;

      delete conditions_;

   }
};


typedef std::list< cxxtls::ThreadSafeHandle<CaseMember> >CaseMemberList;


struct BourneCaseStatement
:  public Statement
{
    auto_ptr<StringExpression>                            testExpression_;  // owned of course

    std::auto_ptr<CaseMemberList>                          cases_;

    std::auto_ptr<RedirType::REDIRECTION_LIST> redirections_;

    BourneCaseStatement(StringExpression*           testExpression,
                        CaseMemberList*              cases,
                        RedirType::REDIRECTION_LIST* redirections
                       )
        /// @param[in,out] testExpression -- Adopted by this class.  Will be deleted at destruction.
        ///                                  This parameter is the word that will be tested against
        ///                                  all the cases.
        /// @param[in,out] cases             Adopted by this class.  Will be deleted at destruction.
        ///                                  This is a list of cases containing patterns and 
        ///                                  statements to execute when the test expression matches
        ///                                  the pattern.
        /// @param[in,out] redirections      Adopted by this class.  Will be deleted at destruction,
        ///                                  This is the command line file redirections for the statement.
    :  testExpression_(testExpression)
    ,  cases_(cases)
    ,  redirections_(redirections)
    {
    }

    virtual ~BourneCaseStatement();

    virtual StatementReturnCode execute(ScriptThread::ThreadContext             *executionContext,
                                        FunctionScope                           *functionScope,
                                        cxxtls::ScriptFunctions::JobController  *jobController,
                                        bool                                          skipArg0            // for use by the command "builtin"
                                       ) const;

    std::string getDebugDescription() const
    {
        return "case";
    }
};


//==========================================================================================================
//==========================================================================================================
//==========================================================================================================

    struct NotStatement
    : public Statement
        ///  Execute another statement and invert the meaning of the error code that it sets.
        ///  If the error code is zero, it will become one.  Otherwise, it will become zero.
    {
        virtual ~NotStatement();

        ThreadSafeHandle<Statement> mainClause_;

        std::auto_ptr<RedirType::REDIRECTION_LIST> redirections_;

        NotStatement(ThreadSafeHandle<Statement> &mainClause)
        :  mainClause_(mainClause)
        {
        }

        virtual StatementReturnCode execute(ScriptThread::ThreadContext             *executionContext,
                                            FunctionScope                           *functionScope,
                                            cxxtls::ScriptFunctions::JobController  *jobController,
                                            bool                                          skipArg0            // for use by the command "builtin"
                                           ) const;

        virtual std::string getDebugDescription() const 
        {
            if(&*mainClause_)
               return std::string("! ") + mainClause_->getDebugDescription();

            return "!";
        }


    };

//==========================================================================================================
//==========================================================================================================
//==========================================================================================================

    struct BourneConditionalStatement
    : public Statement
       /// Either && or || statement
    {
        virtual ~BourneConditionalStatement();

        ThreadSafeHandle<Statement> lhs_;
        ThreadSafeHandle<Statement> rhs_;

        bool isAndAnd_;

        BourneConditionalStatement( ThreadSafeHandle<Statement> &lhs, ThreadSafeHandle<Statement> &rhs, bool isAndAnd)
        :  lhs_(lhs)
        ,  rhs_(rhs)
        ,  isAndAnd_(isAndAnd)
        {
        }

        virtual StatementReturnCode execute(ScriptThread::ThreadContext             *executionContext,
                                            FunctionScope                           *functionScope,
                                            cxxtls::ScriptFunctions::JobController  *jobController,
                                            bool                                          skipArg0            // for use by the command "builtin"
                                           ) const;

        virtual std::string getDebugDescription() const 
        {

            string rv;

            if(&*lhs_)
               rv += lhs_->getDebugDescription();

            rv += ' ';

            if(isAndAnd_)
               rv += "&&";
            else
               rv += "||";

            rv += ' ';

            if(&*rhs_)
               rv += rhs_->getDebugDescription();

            return rv;

        }


    };

//==========================================================================================================
//==========================================================================================================
//==========================================================================================================

    struct BrashWriteExpressions
    : public Statement
       /// Either && or || statement
    {
        virtual ~BrashWriteExpressions();

        std::list<StringExpression*> expressions_;


        BrashWriteExpressions(std::list<std::string> const &expressionTextList)
        {
            CXXTLS_FOREACH(std::string const &cur, expressionTextList)
            {
                string::const_iterator f = cur.begin(),
                                       l = cur.end();

                std::auto_ptr<StringExpression> tmp( cxxtls::StringExpr::parseEmbeddedExpressions(f, l));

                if(tmp.get())
                {
                   expressions_.push_back(tmp.get());
                   tmp.release();
                }
            }
        }
        

        virtual StatementReturnCode execute(ScriptThread::ThreadContext             *executionContext,
                                            FunctionScope                           *functionScope,
                                            cxxtls::ScriptFunctions::JobController  *jobController,
                                            bool                                          skipArg0            // for use by the command "builtin"
                                           ) const;

        std::string getDebugDescription() const 
        {
           return "BrashWriteExpression(???)";
        }


    };

//==========================================================================================================
//==========================================================================================================
//==========================================================================================================

extern std::pair<std::string,bool> findBrashExecutableInPath(std::string const &name, std::string const &PATH);
    ///< Searches for an executable for brash to use in the PATH variable.  On unix, it just looks for a 
    ///  a file in the list of directories found in the PATH variable.  It searches from the beginning of
    ///  the path till the end and stops when it finds a match whose file has the executable bit set --
    ///  and which is not a directory.
    ///
    ///  On windows, though, a different approach is used:  
    ///
    ///  @li the PATH directories are searched in order, and we stop on the first "match"
    ///
    ///  @li a match is defined by one of the following things happening:
    ///
    ///  <blockquote>
    ///      <li>  a matching script name is found in the directory.
    ///            <blockquote>
    ///               <li>  and it is not a sub-directory name
    ///               <li>  A script name is the name of an ascii file whose first line begins with "#!".
    ///            </blockquote>
    ///      <li>  OR, a similar file name which ends in one of the PATHEXT members is found.
    ///  </blockquote>
    ///      
    ///  @return A pair of string, bool.  The string is the name of the file, and the bool is true only if
    ///  the file is expected to be a script rather than an executable.


#endif
